#include <stdio.h>
#include <stddef.h>
#include <stdlib.h>
#include <math.h>
#include "debug.h"
#include "assert.h"
#include "vectorrep.h"
#include "brent.h"
#include "minrisk.h"

#define NAT_AAS 20
#define INFINITY 10000
#define TOLERANCE 0.0001

static double
absolute (double x) {
  return (x >= 0) ? x : -x;
}

static Vector_T
interpolate_new (double alpha, Vector_T phat, Vector_T lambda) {
  int i;
  double beta = 1.0 - alpha, p;
  Vector_T pest = Vector_new(Vector_length(phat));
  
  for (i = 0; i < Vector_length(pest); ++i) {
    p = alpha * Vector_get(phat,i) + beta * Vector_get(lambda,i);
    Vector_put(pest,i,p);
  }
  return pest;
}

/************************************************************************
 *  Squared-Error Background
 ************************************************************************/

static double
argmin_risk_sqerr_background (double *pest, double *lambda, double N) {
  int i;
  double term, sum1 = 0.0, sum2 = 0.0;

  for (i = 0; i < NAT_AAS; ++i) {
    term = pest[i] - lambda[i];
    sum1 += term*term;
    sum2 += pest[i]*pest[i];
  }
  return N*sum1/(N*sum1 + 1.0 - sum2);
}

Vector_T
Minrisk_sqerr_background_iter (Vector_T phat, Vector_T lambda,
			       double **M, double N) {
  double alpha_old, alpha_new;
  Vector_T pest;
  
  alpha_new = N/(N+sqrt(N));
  alpha_old = alpha_new + 2*TOLERANCE;
  while (absolute(alpha_new-alpha_old) > TOLERANCE) {
    printf("%f ",1.0-alpha_new);
    alpha_old = alpha_new;
    pest = interpolate_new(alpha_old,phat,lambda);
    alpha_new = argmin_risk_sqerr_background ((double *) pest->elements,
					      (double *) lambda->elements,N);
    Vector_free(&pest);
  }
  printf("%f\n",1.0-alpha_new);
  return interpolate_new(alpha_new,phat,lambda);
}

Vector_T
Minrisk_sqerr_background_new (Vector_T phat, Vector_T lambda, 
			      double **M, double N) {
  double alpha1, alpha2;
  Vector_T pest1;

  alpha1 = N/(N+sqrt(N));
  pest1 = interpolate_new(alpha1,phat,lambda);
  alpha2 = argmin_risk_sqerr_background((double *) pest1->elements,
					(double *) lambda->elements,N);
  Vector_free(&pest1);
  return interpolate_new(alpha2,phat,lambda);
}



/************************************************************************
 *  Squared-error Subst
 ************************************************************************/

static void
compute_ptilde (double *ptilde, double **M, double *pest) {
  int i, j;
  double sum;

  for (i = 0; i < NAT_AAS; ++i) {
    sum = 0.0;
    for (j = 0; j < NAT_AAS; ++j) {
      sum += M[i][j]*pest[j];
    }
    ptilde[i] = sum;
  }
}

static double
argmin_risk_sqerr_subst (double *pest, double **M, double N) {
  int i, j;
  double sum_diag = 0.0, sum_cross = 0.0, sum_diffsq = 0.0, sum_matrix = 0.0;
  double term, alpha, *ptilde;

  ptilde = (double *) calloc(NAT_AAS,sizeof(double));
  compute_ptilde(ptilde,M,pest);

  for (i = 0; i < NAT_AAS; ++i) {
    term = pest[i] - ptilde[i];
    sum_diffsq += term*term;
    sum_cross += pest[i]*term;
    sum_diag += M[i][i]*pest[i];
    for (j = 0; j < NAT_AAS; ++j) {
      sum_matrix += M[i][j]*M[i][j]*pest[j];
    }
  }
  free(ptilde);
  alpha = (N-1.0)*sum_diffsq + sum_cross - sum_diag + sum_matrix;
  alpha /= 1.0 + (N-1.0)*sum_diffsq - 2.0*sum_diag + sum_matrix;
  return alpha;
}


Vector_T
Minrisk_sqerr_subst_iter (Vector_T phat, Vector_T lambda,
			  double **M, double N) {
  double alpha_old, alpha_new;
  Vector_T pest;
  
  alpha_new = N/(N+sqrt(N));
  alpha_old = alpha_new + 2*TOLERANCE;
  while (absolute(alpha_new-alpha_old) > TOLERANCE) {
    printf("%f ",1.0-alpha_new);
    alpha_old = alpha_new;
    pest = interpolate_new(alpha_old,phat,lambda);
    alpha_new = argmin_risk_sqerr_subst((double *) pest->elements,M,N);
    Vector_free(&pest);
  }
  printf("%f\n",1.0-alpha_new);
  return interpolate_new(alpha_new,phat,lambda);
}


Vector_T
Minrisk_sqerr_subst_new (Vector_T phat, Vector_T lambda, 
			 double **M, double N) {
  double alpha1, alpha2;
  Vector_T pest1;

  alpha1 = N/(N+sqrt(N));
  pest1 = interpolate_new(alpha1,phat,lambda);
  alpha2 = argmin_risk_sqerr_subst((double *) pest1->elements,M,N);
  Vector_free(&pest1);
  return interpolate_new(alpha2,phat,lambda);
}

/************************************************************************
 *  Relative Entropy Background
 ************************************************************************/

static double
rel_ent (double p, double q) {
  if (p == 0.0) {
    return 0.0;
  } else if (q == 0.0) {
    return (double) INFINITY;
  } else {
    return p*log(p/q);
  }
}

static double
risk_relent_background (double alpha, double *phat, double *pest, 
			double *lambda, double **M, double N) {
  int i;
  double pnew, pnew2, p, q, beta;
  double sum1 = 0.0, sum2 = 0.0, sum3 = 0.0, sum4 = 0.0;
  double alpha2 = alpha*alpha;

  beta = 1.0 - alpha;
  for (i = 0; i < NAT_AAS; ++i) {
    p = pest[i];
    q = 1.0 - p;
    pnew = alpha*p + beta*lambda[i];
    pnew2 = pnew*pnew;
    if (pnew == 0.0) {
      sum1 += 0.0;
      sum2 += 1.0/alpha;
      sum3 += -3.0/alpha2;
      sum4 += (2.0-N)/(alpha2*alpha*2.0*N*N*N);
    } else {
      sum1 += rel_ent(pnew,p);
      sum2 += p*q/pnew;
      sum3 += p*q*(q-p)/pnew2;
      sum4 += p*q*(1.0 + 3.0*(N-2.0)*p*q)/(pnew2*pnew);
    }
  }
  sum2 *= alpha2/(2.0*N);
  sum3 *= alpha2*alpha/(6.0*N*N);
  sum4 *= alpha2*alpha2/(12.0*N*N*N);
  debug(printf("%f: %f %f %f %f\n",alpha,sum1,sum2,sum3,sum4));
  return (sum1 + sum2 - sum3 + sum4);
}


Vector_T
Minrisk_relent_background_iter (Vector_T phat, Vector_T lambda,
				double **M, double N) {
  double alpha_old, alpha_new;
  Vector_T pest;
  double ax, bx, cx;
  
  alpha_new = N/(N+sqrt(N));
  alpha_old = alpha_new + 2*TOLERANCE;
  while (absolute(alpha_new-alpha_old) > TOLERANCE) {
    alpha_old = alpha_new;
    pest = interpolate_new(alpha_old,phat,lambda);
    ax = 0.0; bx = 1.0; cx = 0.5;
    alpha_new = Brent(ax,bx,cx,risk_relent_background,
		      (double *) phat->elements,
		      (double *) pest->elements,
		      (double *) lambda->elements,M,N,TOLERANCE);
    Vector_free(&pest);
  }
  return interpolate_new(alpha_new,phat,lambda);
}


Vector_T
Minrisk_relent_background_new (Vector_T phat, Vector_T lambda, 
			       double **M, double N) {
  double alpha1, alpha2;
  Vector_T pest1;
  double ax = 0.0, bx = 1.0, cx = 0.5;

  alpha1 = N/(N+sqrt(N));
  pest1 = interpolate_new(alpha1,phat,lambda);
  alpha2 = Brent(ax,bx,cx,risk_relent_background,
		 (double *) phat->elements,
		 (double *) pest1->elements,
		 (double *) lambda->elements,M,N,TOLERANCE);
  Vector_free(&pest1);
  return interpolate_new(alpha2,phat,lambda);
}


/************************************************************************
 *  Relative Entropy Subst
 ************************************************************************/

static double
risk_relent_subst (double alpha, double *phat, double *pest,
		   double *lambda, double **M, double N) {
  int i;
  double pnew, pnew2, p, q, beta;
  double sum1 = 0.0, sum2 = 0.0, sum3 = 0.0, sum4 = 0.0;
  double adj_alpha, adj_alpha2;
  
  beta = 1.0 - alpha;
  for (i = 0; i < NAT_AAS; ++i) {
    adj_alpha = alpha + beta*M[i][i];
    adj_alpha2 = adj_alpha*adj_alpha;
    p = pest[i];
    q = 1.0 - p;
    pnew = alpha*p + beta*M[i][i]*p + beta*lambda[i] - beta*M[i][i]*phat[i];
    pnew2 = pnew*pnew;
    if (pnew == 0.0) {
      sum1 += 0.0;
      sum2 += 1.0/adj_alpha;
      sum3 += -3.0/adj_alpha2;
      sum4 += (2.0-N)/(adj_alpha2*adj_alpha*2.0*N*N*N);
    } else {
      sum1 += rel_ent(pnew,p);
      sum2 += (p*q/pnew)*(adj_alpha2);
      sum3 += (p*q*(q-p)/pnew2)*(adj_alpha2*adj_alpha);
      sum4 += (p*q*(1.0 + 3.0*(N-2.0)*p*q)/(pnew2*pnew))*adj_alpha2*adj_alpha2;
    }
  }
  sum2 *= 1.0/(2.0*N);
  sum3 *= 1.0/(6.0*N*N);
  sum4 *= 1.0/(12.0*N*N*N);
  debug(printf("%f: %f %f %f %f\n",alpha,sum1,sum2,sum3,sum4));
  return (sum1 + sum2 - sum3 + sum4);
}

Vector_T
Minrisk_relent_subst_iter (Vector_T phat, Vector_T lambda,
			   double **M, double N) {
  double alpha_old, alpha_new;
  Vector_T pest;
  double ax, bx, cx;
  
  alpha_new = N/(N+sqrt(N));
  alpha_old = alpha_new + 2*TOLERANCE;
  while (absolute(alpha_new-alpha_old) > TOLERANCE) {
    alpha_old = alpha_new;
    pest = interpolate_new(alpha_old,phat,lambda);
    ax = 0.0; bx = 1.0; cx = 0.5;
    alpha_new = Brent(ax,bx,cx,risk_relent_subst,
		      (double *) phat->elements,
		      (double *) pest->elements,
		      (double *) lambda->elements,M,N,TOLERANCE);
    Vector_free(&pest);
  }
  return interpolate_new(alpha_new,phat,lambda);
}

Vector_T
Minrisk_relent_subst_new (Vector_T phat, Vector_T lambda, double **M, double N) {
  double alpha1, alpha2;
  Vector_T pest1;
  double ax = 0.0, bx = 1.0, cx = 0.5;

  alpha1 = N/(N+sqrt(N));
  pest1 = interpolate_new(alpha1,phat,lambda);
  alpha2 = Brent(ax,bx,cx,risk_relent_subst,
		 (double *) phat->elements,
		 (double *) pest1->elements,
		 (double *) lambda->elements,M,N,TOLERANCE);
  Vector_free(&pest1);
  return interpolate_new(alpha2,phat,lambda);
}





